@@ -3,7 +3,6 @@
 # This source code is licensed under the MIT license found in the
 # LICENSE file in the root directory of this source tree.
 
-# pyre-strict
 
 import abc
 import builtins
@@ -14,7 +13,6 @@ from enum import Enum, auto
 from typing import (
     Collection,
     Dict,
-    Iterable,
     Iterator,
     List,
     Mapping,
@@ -48,7 +46,7 @@ class Access:
        This scope analysis only analyzes access via a :class:`~libcst.Name` or  a :class:`~libcst.Name`
        node embedded in other node like :class:`~libcst.Call` or :class:`~libcst.Attribute`.
        It doesn't support type annontation using :class:`~libcst.SimpleString` literal for forward
-       references. E.g. in this example, the ``"Tree"`` isn't parsed as as an access::
+       references. E.g. in this example, the ``"Tree"`` isn't parsed as an access::
 
            class Tree:
                def __new__(cls) -> "Tree":
@@ -65,11 +63,14 @@ class Access:
     #: assignment.
     scope: "Scope"
 
+    is_annotation: bool
+
     __assignments: Set["BaseAssignment"]
 
-    def __init__(self, node: cst.Name, scope: "Scope") -> None:
+    def __init__(self, node: cst.Name, scope: "Scope", is_annotation: bool) -> None:
         self.node = node
         self.scope = scope
+        self.is_annotation = is_annotation
         self.__assignments = set()
 
     def __hash__(self) -> int:
@@ -249,7 +250,7 @@ class _NameUtil:
                         if eval_alias is not None:
                             as_name = eval_alias
                     if full_name.startswith(as_name):
-                        remaining_name = full_name.split(as_name)[1].lstrip(".")
+                        remaining_name = full_name.split(as_name, 1)[1].lstrip(".")
                         results.add(
                             QualifiedName(
                                 f"{real_name}.{remaining_name}"
@@ -607,19 +608,32 @@ class ComprehensionScope(LocalScope):
 # each string has the corresponding CSTNode attached to it
 def _gen_dotted_names(
     node: Union[cst.Attribute, cst.Name]
-) -> Iterable[Tuple[str, Union[cst.Attribute, cst.Name]]]:
+) -> Iterator[Tuple[str, Union[cst.Attribute, cst.Name]]]:
     if isinstance(node, cst.Name):
-        yield (node.value, node)
+        yield node.value, node
     else:
         value = node.value
-        if not isinstance(value, (cst.Attribute, cst.Name)):
-            # this is not an import
-            return
-        name_values = iter(_gen_dotted_names(value))
-        (next_name, next_node) = next(name_values)
-        yield (f"{next_name}.{node.attr.value}", node)
-        yield (next_name, next_node)
-        yield from name_values
+        if isinstance(value, cst.Call):
+            value = value.func
+            if isinstance(value, (cst.Attribute, cst.Name)):
+                name_values = _gen_dotted_names(value)
+                try:
+                    next_name, next_node = next(name_values)
+                except StopIteration:
+                    return
+                else:
+                    yield next_name, next_node
+                    yield from name_values
+        elif isinstance(value, (cst.Attribute, cst.Name)):
+            name_values = _gen_dotted_names(value)
+            try:
+                next_name, next_node = next(name_values)
+            except StopIteration:
+                return
+            else:
+                yield f"{next_name}.{node.attr.value}", node
+                yield next_name, next_node
+                yield from name_values
 
 
 class ScopeVisitor(cst.CSTVisitor):
@@ -628,7 +642,8 @@ class ScopeVisitor(cst.CSTVisitor):
         self.provider: ScopeProvider = provider
         self.scope: Scope = GlobalScope()
         self.__deferred_accesses: List[Tuple[Access, Optional[cst.Attribute]]] = []
-        self.__top_level_attribute: Optional[cst.Attribute] = None
+        self.__top_level_attribute_stack: List[Optional[cst.Attribute]] = [None]
+        self.__in_annotation: Set[Union[cst.Call, cst.Annotation]] = set()
 
     @contextmanager
     def _new_scope(
@@ -657,6 +672,7 @@ class ScopeVisitor(cst.CSTVisitor):
 
         # make sure node.names is Sequence[ImportAlias]
         for name in names:
+            self.provider.set_metadata(name, self.scope)
             asname = name.asname
             if asname is not None:
                 name_values = _gen_dotted_names(cst.ensure_type(asname.name, cst.Name))
@@ -674,21 +690,72 @@ class ScopeVisitor(cst.CSTVisitor):
         return self._visit_import_alike(node)
 
     def visit_Attribute(self, node: cst.Attribute) -> Optional[bool]:
-        if self.__top_level_attribute is None:
-            self.__top_level_attribute = node
+        if self.__top_level_attribute_stack[-1] is None:
+            self.__top_level_attribute_stack[-1] = node
         node.value.visit(self)  # explicitly not visiting attr
-        if self.__top_level_attribute is node:
-            self.__top_level_attribute = None
+        if self.__top_level_attribute_stack[-1] is node:
+            self.__top_level_attribute_stack[-1] = None
         return False
 
+    def visit_Call(self, node: cst.Call) -> Optional[bool]:
+        self.__top_level_attribute_stack.append(None)
+        if any(
+            qn.name == "typing.TypeVar"
+            for qn in self.scope.get_qualified_names_for(node)
+        ):
+            node.func.visit(self)
+            self.__in_annotation.add(node)
+            for arg in node.args[1:]:
+                arg.visit(self)
+            return False
+        return True
+
+    def leave_Call(self, original_node: cst.Call) -> None:
+        self.__top_level_attribute_stack.pop()
+        self.__in_annotation.discard(original_node)
+
+    def visit_Annotation(self, node: cst.Annotation) -> Optional[bool]:
+        self.__in_annotation.add(node)
+
+    def leave_Annotation(self, original_node: cst.Annotation) -> None:
+        self.__in_annotation.discard(original_node)
+
+    def visit_SimpleString(self, node: cst.SimpleString) -> Optional[bool]:
+        self._handle_string_annotation(node)
+        return False
+
+    def visit_ConcatenatedString(self, node: cst.ConcatenatedString) -> Optional[bool]:
+        self._handle_string_annotation(node)
+        return False
+
+    def _handle_string_annotation(
+        self, node: Union[cst.SimpleString, cst.ConcatenatedString]
+    ) -> None:
+        if self.__in_annotation:
+            value = node.evaluated_value
+            if value:
+                mod = cst.parse_module(value)
+                mod.visit(self)
+
+    def visit_Subscript(self, node: cst.Subscript) -> Optional[bool]:
+        if any(
+            qn.name in ("typing.Literal", "typing_extensions.Literal")
+            for qn in self.scope.get_qualified_names_for(node.value)
+        ):
+            node.value.visit(self)
+            return False
+        return True
+
     def visit_Name(self, node: cst.Name) -> Optional[bool]:
         # not all Name have ExpressionContext
         context = self.provider.get_metadata(ExpressionContextProvider, node, None)
         if context == ExpressionContext.STORE:
             self.scope.record_assignment(node.value, node)
-        elif context in (ExpressionContext.LOAD, ExpressionContext.DEL):
-            access = Access(node, self.scope)
-            self.__deferred_accesses.append((access, self.__top_level_attribute))
+        elif context in (ExpressionContext.LOAD, ExpressionContext.DEL, None):
+            access = Access(node, self.scope, is_annotation=bool(self.__in_annotation))
+            self.__deferred_accesses.append(
+                (access, self.__top_level_attribute_stack[-1])
+            )
 
     def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
         self.scope.record_assignment(node.name.value, node)
@@ -731,6 +798,7 @@ class ScopeVisitor(cst.CSTVisitor):
 
     def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
         self.scope.record_assignment(node.name.value, node)
+        self.provider.set_metadata(node.name, self.scope)
         for decorator in node.decorators:
             decorator.visit(self)
         for base in node.bases:
@@ -825,16 +893,16 @@ class ScopeVisitor(cst.CSTVisitor):
         # the time complexity is O(m x n), this optimizes it as O(m + n).
         scope_name_accesses = defaultdict(set)
         for (access, enclosing_attribute) in self.__deferred_accesses:
+            name = ensure_type(access.node, cst.Name).value
             if enclosing_attribute is not None:
-                name = None
-                for name, node in _gen_dotted_names(enclosing_attribute):
-                    if name in access.scope:
+                # if _gen_dotted_names doesn't generate any values, fall back to
+                # the original name node above
+                for attr_name, node in _gen_dotted_names(enclosing_attribute):
+                    if attr_name in access.scope:
                         access.node = node
+                        name = attr_name
                         break
-                if name is None:
-                    continue
-            else:
-                name = ensure_type(access.node, cst.Name).value
+
             scope_name_accesses[(access.scope, name)].add(access)
             access.record_assignments(access.scope[name])
             access.scope.record_access(name, access)
